from puffmatic import response
from unittest import mock

from os.path import exists, join
from os import makedirs

import filecmp
import pytest


@mock.patch("puffmatic.utils.run")
def test_run(run):
    response._run(["alfa", "1", "2"])
    assert run.call_count == 1
    assert run.call_args.args == (["alfa", "1", "2"], )
    assert run.call_args.kwargs.get("logger") is not None
    assert run.call_args.kwargs.get("check") is True


class TestRenderInstallResp():

    def test_copy_hardcoded_response_file(self, cfg, hosts_dir, tmpdir):
        # GIVEN
        #     host configuration has hardcoded autoinstall config
        #     config has trailing newline
        #     output directory exists
        autoinstall_file = join(hosts_dir, "example", "install.conf")
        assert exists(autoinstall_file)
        with open(autoinstall_file) as f:
            assert f.read()[-1] == '\n'
        autoinstall_output_file = join(cfg.resp_output_dir, "install.conf")
        makedirs(cfg.resp_output_dir)

        # WHEN
        #     config rendered
        response.render_install_resp(install_resp_template=autoinstall_file,
                                     output_file=autoinstall_output_file,
                                     context=vars(cfg))

        # THEN
        #     file is copied
        #     trailing whitespace is retained
        assert filecmp.cmp(autoinstall_file, autoinstall_output_file)

    def test_render_from_template(self, cfg, templates_dir, tmpdir):
        # GIVEN
        #     host has autoinstall jinja template
        template_file = join(templates_dir, "install.conf.j2")
        assert exists(template_file)
        output_file = join(cfg.resp_output_dir, "auto_install.conf")

        # WHEN
        #     response rendered
        response.render_install_resp(install_resp_template=template_file,
                                     output_file=output_file,
                                     context=vars(cfg))

        # THEN
        #     template filled with data from host config
        with open(output_file, "r") as f:
            rendered = f.read()
        assert len(rendered) > 0
        assert f"System hostname = {cfg.hostname}" in rendered


class TestResponseMain():

    def test_host_harcoded_response(self, cfg, hosts_dir):
        host_dir = join(hosts_dir, "example")
        response.response_main(cfg, host_dir)
        output_file = join(cfg.resp_output_dir, "install-example.conf")
        assert exists(output_file)
        with open(output_file) as f:
            rendered = f.read()
            assert "System hostname = example" in rendered
            assert "Host hardcoded = yes" in rendered

    def test_host_template_response(self, cfg, hosts_dir):
        # GIVEN
        #     host config has response template
        cfg.hostname = "bravo"
        host_dir = join(hosts_dir, "bravo")
        host_template = join(host_dir, "install.conf.j2")
        assert exists(host_template)

        # WHEN
        #     autoinstall config rendered
        response.response_main(cfg, host_dir)

        # THEN
        #     file placed in response file directory
        #     template filled with data from host config
        #     template came from host config directory
        output_file = join(cfg.resp_output_dir, "install-bravo.conf")
        assert exists(output_file)
        with open(output_file) as f:
            rendered = f.read()
            assert "System hostname = bravo" in rendered
            assert "Bravo host template = yes" in rendered

    def test_use_generic_response(self, cfg, hosts_dir):
        # GIVEN
        #     host configuration without resp or template
        #     default template exists
        cfg.hostname = "charlie"
        host_dir = join(hosts_dir, "charlie")

        # WHEN
        #     response rendered
        response.response_main(cfg, host_dir)
        output_file = join(cfg.resp_output_dir, "install-charlie.conf")

        # THEN
        #     file rendered from default template
        assert exists(output_file)
        with open(output_file) as f:
            rendered = f.read()
            assert "System hostname = charlie" in rendered
            assert "Default template = yes" in rendered

    def test_missing_response_template(self, cfg, hosts_dir, tmpdir):
        # GIVEN
        #     no response file
        #     no respones file template
        cfg.hostname = "empty"
        cfg.templates_dir = join(tmpdir, "empty")
        empty_host_dir = join(hosts_dir, "empty")

        # WHEN
        #     response rendered
        with pytest.raises(SystemExit) as e:
            response.response_main(cfg, empty_host_dir)

        # THEN
        #     exit with error code
        assert e.value.code == 1



import tarfile

from puffmatic import sets
from unittest import mock

from os.path import basename, dirname, join, exists


@mock.patch("puffmatic.utils.run")
def test_run(run):
    sets._run(["alfa", "1", "2"])
    assert run.call_count == 1
    assert run.call_args.args == (["alfa", "1", "2"], )
    assert run.call_args.kwargs.get("logger") is not None
    assert run.call_args.kwargs.get("check") is True


@mock.patch("puffmatic.utils.run")
def test_test_mount_tmp_dir(run, tmpdir):
    sets.mount_tmp_dir(tmp_dir=tmpdir, size=128)
    assert run.call_args.kwargs.get("priv") is True
    assert any("-s128M" in str(arg) for arg in run.call_args.args[0])


@mock.patch("puffmatic.utils.run")
def test_umount_tmp_dir(run, tmpdir):
    sets.umount_tmp_dir(tmp_dir=tmpdir)
    assert run.call_args.kwargs.get("priv") is True


@mock.patch("puffmatic.utils.run")
def test_run_mtree(run, tmpdir):
    site_dir = tmpdir / "site"
    mtree_file = tmpdir / "site.mtree"
    sets.run_mtree(site_dir=site_dir, mtree_file=mtree_file)
    assert run.call_args.kwargs.get("cwd") == site_dir


@mock.patch("puffmatic.utils.run")
def test_create_tarball(run, tmpdir):
    site_dir = tmpdir / "site"
    output_file = tmpdir / "arch.tgz"
    sets.create_tarball(site_dir, output_file)
    assert run.call_args.kwargs.get("cwd") == site_dir


def test_update_index(tmpdir):
    import pathlib
    pathlib.Path(tmpdir / "alfa").touch()
    pathlib.Path(tmpdir / "bravo").touch()
    sets.update_index(tmpdir)
    with open(tmpdir / "index.txt") as f:
        index = f.read()
        assert "alfa" in index
        assert "bravo" in index


def test_create_site_set(cfg, hosts_dir, tmpdir):
    # GIVEN
    #     site dir with mtree spec
    site_dir = join(hosts_dir, "example", "site")
    mtree_file = join(hosts_dir, "example", "site.mtree")

    # WHEN
    #     site archive is created
    sets.create_site_set(site_dir=site_dir,
                         mtree_file=mtree_file,
                         output_dir=cfg.sets_output_dir,
                         archive="site77.tgz",
                         tmp_dir=tmpdir)
    site_archive = join(cfg.sets_output_dir, "site77.tgz")

    # THEN
    #     archive is placed in sets directory
    #     archive is compressed
    #     it contains all files
    #     mode modified before archiving
    assert exists(site_archive)
    with tarfile.open(site_archive, "r:gz") as tar:
        members = tar.getmembers()
        files = [m.name for m in members]
        assert "./etc/ssh/ssh_host_rsa_key" in files
        assert "./etc/ssh/ssh_host_rsa_key.pub" in files
        assert "./etc/ssh/ssh_host_ecdsa_key" in files
        assert "./etc/ssh/ssh_host_ecdsa_key.pub" in files
        assert "./etc/ssh/ssh_host_ed25519_key" in files
        assert "./etc/ssh/ssh_host_ed25519_key.pub" in files

        for m in members:
            if m.name.endswith("_key"):
                assert (m.mode & 0o777) == 0o600
            if m.name.endswith("_key.pub"):
                assert (m.mode & 0o777) == 0o644


class TestHostSiteSetMain():

    def test_no_host_dir(self, cfg, tmpdir):
        # GIVEN
        #     host site directory does not exist
        import pytest
        cfg.hostname = "non-existing-host"
        non_existing_host_dir = join(tmpdir, "non-existing-host")

        # WHEN
        #     attempting to create host site set
        with pytest.raises(SystemExit) as se:
            sets.host_site_set_main(cfg=cfg,
                                    host_dir=non_existing_host_dir)

        # THEN
        #     no action
        #     graceful exit
        assert se.value.code == 0

    def test_set_file_created(self, cfg, hosts_dir, tmpdir):
        # GIVEN
        #     host exists with site dir
        cfg.hostname = "example"
        host_dir = join(hosts_dir, "example")

        # WHEN
        #     creating host site set
        sets.host_site_set_main(cfg=cfg, host_dir=host_dir)

        # THEN
        #     archive is placed in output directory
        #     host suffix
        #     version in short format
        set_archive = f"site{cfg.version_short}-{cfg.hostname}.tgz"
        set_archive_path = join(cfg.sets_output_dir, set_archive)
        assert exists(set_archive_path)


def test_site_set_main(cfg, hosts_dir, tmpdir):
    # GIVEN
    #     domain site dir exists
    site_dir = join(hosts_dir, "example", "site")

    # WHEN
    #     creating site set
    sets.site_set_main(cfg=cfg,
                       site_dir=site_dir)

    # THEN
    #     site archive created
    #     no hostname suffx
    #     version in short format
    site_archive_path = join(cfg.sets_output_dir,
                             f"site{cfg.version_short}.tgz")
    assert exists(site_archive_path)



"""Provide fixtures for all tests."""
import os
import pytest
import subprocess
import puffmatic.utils


__dir = os.path.dirname(__file__)
__data_dir = os.path.join(__dir, "data")
__fixtures_dir = os.path.join(__dir, "fixtures")


def pytest_addoption(parser):
    """Configure custom options to enable custom test behavior."""
    parser.addoption("--run-integration",
                     action="store_true",
                     default=False,
                     help="run integration tests against real resources")
    parser.addoption("--work-dir",
                     action="store",
                     default="/release/puffmatic/",
                     help="tests working directory")


def pytest_configure(config):
    """Configure pytest runner."""
    basetemp = os.path.join(config.option.work_dir, "tmp")
    config.option.basetemp = basetemp


@pytest.fixture
def work_dir(pytestconfig):
    """
    Test suite working directory.

    Working directory contains temp directory as well
    as dynamically generated fixture data.
    """
    return pytestconfig.option.work_dir


@pytest.fixture
def mirror_dir2(work_dir, cfg):  # pragma: nocover
    """
    Directory containing mirrored OpenBSD release.

    This fixture lives outside temporary directory so it will
    be persisted between test run invocation. Mirroring entire
    release is time consuming and require several GBs of disk space.
    """
    mirror_dir = os.path.join(work_dir, "mirror", "OpenBSD",
                              cfg.version_full, cfg.arch) + os.sep
    stamp_file = os.path.join(mirror_dir, ".stamp")
    if os.path.isfile(stamp_file):
        return mirror_dir
    src = ("rsync://mirror.planetunix.net/OpenBSD/"
           f"{cfg.version_full}/{cfg.arch}")
    os.makedirs(mirror_dir, exist_ok=True)
    assert mirror_dir.endswith("/")
    subprocess.run(["openrsync", "-v", "-r", src, mirror_dir],
                   cwd=mirror_dir,
                   check=True)
    with open(stamp_file, "w") as f:
        f.write("stamp")
    return mirror_dir


@pytest.fixture
def config_file():
    config_yaml = os.path.join(__fixtures_dir, "config", "config.yaml")
    assert os.path.isfile(config_yaml)
    return config_yaml


@pytest.fixture
def cfg(tmpdir, work_dir):
    """
    Load generator config fixture.

    Output directory resides in temporary directory.
    Temporary directory will be deleted after test run.

    Templates directory points to fixture directory. Do not
    modify templates directory.
    """
    config_yaml = os.path.join(__fixtures_dir, "config", "config.yaml")
    assert os.path.exists(config_yaml)
    cfg = puffmatic.utils.Config.load(config_yaml)
    cfg.hostname = "example"
    cfg.output_dir = os.path.join(tmpdir, "output")
    os.makedirs(cfg.output_dir, exist_ok=True)
    cfg.templates_dir = os.path.join(__fixtures_dir, "config", "templates")
    mirror_dir = os.path.join(work_dir, "mirror", "OpenBSD")
    cfg.mirror = mirror_dir
    return cfg


@pytest.fixture
def hosts_dir():
    """Return path to host configurations."""
    return os.path.join(__fixtures_dir, "config", "hosts")


@pytest.fixture
def templates_dir():
    """Return path to test templates directory."""
    return os.path.join(__fixtures_dir, "config", "templates")



"""Main."""
import argparse
import logging
import os.path
import sys

from puffmatic.disklabel import disklabel_main
from puffmatic.mirror import mirror_main
from puffmatic.response import response_main
from puffmatic.sets import host_site_set_main, site_set_main
from puffmatic.getimg import getimg_main
from puffmatic.patchimg import patchimg_main, patchimg_umount_main
from puffmatic.utils import Config, create_logger


logger = create_logger()


def load_config(config_file="config.yaml") -> Config:
    if config_file:
        return Config.load(config_file)
    elif os.path.exists("config.yaml"):
        return Config.load("config.yaml")
    else:
        raise RuntimeError("no config file provided")


def main(args):
    """Execute main body."""
    args_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description='Process some commands.',
        formatter_class=args_formatter,
        prog="puffmatic"
    )
    parser.add_argument("--verbose", "-v", action="store_true")
    parser.add_argument("--config", "-c", type=str)
    subpar = parser.add_subparsers(dest='command', required=True)

    _ = subpar.add_parser("help", help="Display help")

    mirror_parser = subpar.add_parser("mirror",
                                      formatter_class=args_formatter,
                                      help="Mirror OpenBSD release directory")

    mirror_parser.add_argument("--once", "-1",
                               action="store_true",
                               help=("Mirror only once and stamp directory. "
                                     "Do not re-download if stamp exists."))

    host_site_set_parser = subpar.add_parser("host-site-set",
                                             formatter_class=args_formatter,
                                             help="Create host site set archive.")
    host_site_set_parser.add_argument("--host-dir", "-H",
                                      type=str,
                                      required=True,
                                      help="Directory with host site set content")

    site_set_parser = subpar.add_parser("site-set",
                                        formatter_class=args_formatter,
                                        help="Create site set archive.")
    site_set_parser.add_argument("--site-dir", "-d",
                                 type=str,
                                 required=True,
                                 help="Directory with common site set content")

    disklabel_parser = subpar.add_parser("disklabel",
                                         formatter_class=args_formatter,
                                         help="Generate disklabel file")
    disklabel_parser.add_argument("--host-dir", "-H",
                                  type=str,
                                  required=True,
                                  help="directory with host configuration")

    response_parser = subpar.add_parser("response",
                                        formatter_class=args_formatter,
                                        help="Generate response file")
    response_parser.add_argument("--host-dir", "-H",
                                 type=str,
                                 required=True,
                                 help="directory with host configuration")

    getimg_parser = subpar.add_parser("getimg",
                                      formatter_class=args_formatter,
                                      help="Download installer image")
    getimg_parser.add_argument("--once", "-1",
                               action="store_true",
                               help=("Download image once and stamp it. "
                                     "Do not re-download if stamp exists."))

    patchimg_parser = subpar.add_parser("patchimg",
                                        formatter_class=args_formatter,
                                        help="Patch installer image")
    patchimg_parser.add_argument("--host-dir", "-H",
                                 type=str,
                                 required=True,
                                 help="directory with host configuration")

    _ = subpar.add_parser("patchimg-umount",
                          help="umount any installer image left by patchimg")

    _ = subpar.add_parser("env", help="print config and environment")

    args = parser.parse_args(args=args)
    kwargs = vars(args)
    if args.verbose:
        logger.setLevel(logging.DEBUG)
        logger.debug("debug is enabled")
    if args.command == "help":
        parser.print_help()
        return 0

    try:
        cfg = load_config(args.config)
    except BaseException as e:
        logger.error(e)
        return 1

    if args.command == "mirror":
        mirror_main(cfg, **vars(args))
    elif args.command == "host-site-set":
        host_site_set_main(cfg, **vars(args))
    elif args.command == "site-set":
        site_set_main(cfg, **vars(args))
    elif args.command == "disklabel":
        disklabel_main(cfg, **vars(args))
    elif args.command == "response":
        response_main(cfg, **vars(args))
    elif args.command == "getimg":
        getimg_main(cfg, **vars(args))
    elif args.command == "patchimg":
        patchimg_main(cfg, **kwargs)
    elif args.command == "patchimg-umount":
        patchimg_umount_main(cfg)
    else:
        print(vars(cfg))
        return 0


if __name__ == "__main__":  # pragma: nocover
    rc = main(args=sys.argv[1:])
    sys.exit(rc)



"""Disklabel generator."""

import logging

from os import makedirs
from os.path import (abspath,
                     exists,
                     basename,
                     join)
from shutil import copyfile
from sys import exit


logger = logging.getLogger().getChild(__name__)


def disklabel_main(cfg, host_dir, **kwargs):
    """Copy disklabel file to installation media directory."""
    host_dir = abspath(host_dir)
    output_dir = abspath(cfg.resp_output_dir)
    templates_dir = abspath(cfg.templates_dir)
    hostname = basename(host_dir)
    host_disklabel_file = join(host_dir, "disklabel")
    generic_disklabel_file = join(templates_dir, "disklabel")
    output_disklabel_file = join(output_dir, f"disklabel-{hostname}")
    makedirs(output_dir, exist_ok=True)
    if exists(host_disklabel_file):
        logger.info((f"copying {host_disklabel_file} "
                     f"to {output_disklabel_file}"))
        copyfile(host_disklabel_file, output_disklabel_file)
    elif exists(generic_disklabel_file):
        logger.info((f"copying {generic_disklabel_file} "
                     f"to {output_disklabel_file}"))
        copyfile(generic_disklabel_file, output_disklabel_file)
    else:
        logger.error("no disklabel file or template found")
        exit(1)



import logging

from puffmatic import mirror
from os import listdir, makedirs
from os.path import join
from unittest import mock

logger = logging.getLogger().getChild(__name__)


def test_mirror_release_dir(cfg, mirror_dir2):
    # GIVEN
    #     files stored on mirror server
    mirror_files = set(listdir(mirror_dir2))

    # WHEN
    #     files are downloaded to output dir
    mirror.mirror_release_dir(version=cfg.version_full,
                              arch=cfg.arch,
                              output_dir=cfg.output_dir,
                              mirror=cfg.mirror)

    # THEN
    #     all files from remote directory downloaded
    downloaded_files = set(listdir(cfg.output_dir))
    assert downloaded_files == mirror_files


def test_check_release_dir(cfg, mirror_dir2):
    """
    Check reference mirror fixture directory.

    Since the mirrored directory is a golden sample,
    check should pass.
    """
    assert mirror.check_release_dir(mirror_dir2) == 0


class TestMirrorMain():

    def test_mirror_main(self, cfg, mirror_dir2):
        # GIVEN
        #     mirror with openbsd release
        mirror_files = set(listdir(mirror_dir2))

        # WHEN
        #     release files mirrored into local directory
        mirror.mirror_main(cfg=cfg, once=False)

        # THEN
        #     all files downloaded
        #     integrity check passed
        #     stamp created
        downloaded_files = set(listdir(cfg.sets_output_dir))
        assert mirror_files.issubset(downloaded_files)
        assert "mirror.stamp" in downloaded_files

    @mock.patch("puffmatic.mirror.mirror_release_dir")
    def test_skip_download_if_stamp_exists(self, mirror_release_dir, cfg):
        # GIVEN
        #     mirror stamp exists
        stamp_file = join(cfg.sets_output_dir, "mirror.stamp")
        makedirs(cfg.sets_output_dir, exist_ok=True)
        with open(stamp_file, "w") as f:
            f.write("test stamp")

        # WHEN
        #     mirror with "once" option
        rc = mirror.mirror_main(cfg=cfg, once=True)

        # THEN
        #     download skipped
        assert rc == 0
        assert mirror_release_dir.call_count == 0



from . import fixture_file
from os.path import isfile
import pytest


class TestFixtureFile():

    def test_file_found(self):
        p = fixture_file("bin.txt")
        assert isfile(p)

    def test_file_not_found(self):
        with pytest.raises(AssertionError):
            _ = fixture_file("does-not-exist")



import logging
import pytest
import subprocess

from puffmatic import utils
from puffmatic.tests import fixture_file
from os.path import (abspath, dirname, join)
from unittest import mock
from urllib.parse import urlparse


class TestCreateLogger():

    def test_root_logger(self):
        logger = utils.create_logger()
        assert logger.name == "root"

    def test_create_named_logger(self):
        logger = utils.create_logger("a.b.c")
        assert logger.name == "a.b.c"


class TestCmdOutputLogger():

    stdout_lines = ["stdout 1", "stdout 2", "stdout 3"]
    stderr_lines = ["stderr 1", "stderr 2", "stderr 3", "stderr 4"]

    @pytest.fixture
    def logger(self):
        return mock.Mock(spec=logging.Logger)

    @pytest.fixture
    def stdout_file(self, tmp_path):
        stdout_file = tmp_path / "stdout.txt"
        stdout_file.write_text("\n".join(self.stdout_lines))
        return stdout_file

    @pytest.fixture
    def stderr_file(self, tmp_path):
        stdout_file = tmp_path / "stderr.txt"
        stdout_file.write_text("\n".join(self.stderr_lines))
        return stdout_file

    def test_no_logger(self, stdout_file):
        with open(stdout_file, "r") as f:
            utils._cmd_output_logger(logger=None, stdout=stdout_file)

    def test_reads_all_data(self, logger, stdout_file, stderr_file):
        all_output_lines = self.stdout_lines + self.stderr_lines
        with (open(stdout_file, "r") as stdout,
              open(stderr_file, "r") as stderr):
            utils._cmd_output_logger(logger=logger,
                                     stdout=stdout,
                                     stderr=stderr)
        assert logger.info.call_count == len(all_output_lines)
        expected_info_calls = [mock.call(line) for line in all_output_lines]
        logger.info.assert_has_calls(expected_info_calls, any_order=True)


class TestFormatCmd():

    def test_command_and_args(self):
        r = utils._format_cmd(["foo", "arg1", "arg2"])
        assert r == "foo arg1 arg2"

    def test_stdout_redirection(self):
        with open("/dev/null", "w") as f:
            r = utils._format_cmd(["foo", "arg1", "arg2"], stdout=f)
        assert r == "foo arg1 arg2 > /dev/null"

    def test_stdin_redirection(self):
        with open("/dev/null", "r") as f:
            r = utils._format_cmd(["foo", "arg1", "arg2"], stdin=f)
        assert r == "foo arg1 arg2 < /dev/null"

    def test_stderr_redirection(self):
        with open("/dev/null", "r") as f:
            r = utils._format_cmd(["foo", "arg1", "arg2"], stderr=f)
        assert r == "foo arg1 arg2 2> /dev/null"


class TestParseVersion():

    def test_valid_version(self):
        v = utils.parse_version("7.6")
        assert v == (7, 6)

    def test_not_a_version(self):
        invalid_versions = ["abc",
                            "a.b",
                            "7.",
                            "7",
                            ".",
                            "..",
                            "1.2.3",
                            "7.a",
                            "a.7"]
        for v in invalid_versions:
            assert not utils.parse_version(v)

    def test_parsing_version_tuple(self):
        v = utils.parse_version((7, 6))
        assert v == (7, 6)


class TestRun():

    @pytest.fixture
    def run_sh(self):
        return fixture_file("utils/run.sh")

    @pytest.fixture
    def logger(self):
        logger = mock.Mock(spec=logging.Logger)
        logger.getChild.return_value = logger
        return logger

    def test_run_captures_stdout(self, run_sh, logger):
        rc = utils.run([run_sh, "stdout"], logger=logger)
        logger.getChild.assert_called_with("run_sh")
        assert rc == 0
        assert logger.info.call_count == 2

    def test_run_captures_stdout_stderr(self, run_sh, logger):
        rc = utils.run([run_sh, "stdout+stderr"], logger=logger)
        logger.getChild.assert_called_with("run_sh")
        assert rc == 0
        assert logger.info.call_count == 4

    def test_return_exit_value(self, run_sh, logger):
        rc = utils.run([run_sh, "fail"], logger=logger)
        assert rc == 1

    def test_raise_when_command_fails(self, run_sh, logger):
        with pytest.raises(subprocess.CalledProcessError):
            utils.run([run_sh, "fail"], logger=logger, check=True)

    def test_no_logger(self, run_sh):
        rc = utils.run([run_sh, "fail"])
        assert rc == 1


def test_load_yaml():
    a = fixture_file("utils/a.yaml")
    b = fixture_file("utils/b.yaml")
    value = utils.load_yaml(a, b)
    assert value == dict(alfa=1,
                         bravo=2,
                         charlie=4,
                         delta=5,
                         echo=6,
                         foxtrot=7)


class TestConfig():

    @pytest.fixture
    def config(self):
        config1_file = fixture_file("utils/config1.yaml")
        config1 = utils.Config.load(config1_file)
        return config1

    def test_load(self, config):
        assert config.version == (7, 6)
        assert config.arch == "amd64"
        assert config.mirror == "rsync://mirror.planetunix.net"
        assert config.install_server == "http://install.example.com"
        assert config.install_basic_auth == "user:password"
        assert config.install_server_dir == "sets"
        assert config.templates_dir == abspath("templates")
        assert config.output_dir == abspath("output")
        assert config.hostname is None
        assert config.ssh_public_key is None
        assert config.root_password is None
        assert config.disk_encryption_passphrase is None
        assert config.ipv6_address is None
        assert config.ipv6_gateway is None

    def test_update(self, config):
        config2_file = fixture_file("utils/config2.yaml")
        config = config.update(config2_file)
        assert config.version == (7, 6)
        assert config.arch == "amd64"
        assert config.mirror == "rsync://mirror.planetunix.net"
        assert config.install_server == "http://install.example.com"
        assert config.install_basic_auth == "user:password"
        assert config.install_server_dir == "sets"
        assert config.templates_dir == abspath("templates")
        assert config.output_dir == abspath("output")
        assert config.hostname == "test-hostname"
        assert config.ssh_public_key == "test-ssh-key"
        assert config.root_password == "test-password"
        assert config.disk_encryption_passphrase == "test-passphrase"
        assert config.ipv6_address == "2001:db8::1"
        assert config.ipv6_gateway == "fe80::1%vio0"

    def test_load_fails(self):
        non_existing_file = join(dirname(__file__), "does-not-exist")
        with pytest.raises(FileNotFoundError):
            utils.Config.load(non_existing_file)

    def test_version_full_str(self, config):
        assert config.version_full == "7.6"

    def test_version_short_str(self, config):
        assert config.version_short == "76"

    def test_resp_server_url(self, config: utils.Config):
        u = urlparse(config.install_server)
        assert f"{u.scheme}://{config.install_basic_auth}@" in config.resp_server_url
        assert u.netloc in config.resp_server_url

    def test_resp_server_url_without_basic_auth(self, config):
        config.install_basic_auth = ""
        assert "@" not in config.resp_server_url

    def test_to_dict(self, config):
        d = config.to_dict()
        assert d["version_full"] == "7.6"
        assert d["version_short"] == "76"
        assert d["sets_output_dir"] == config.sets_output_dir
        assert d["resp_output_dir"] == config.resp_output_dir
        assert d["resp_server_url"] == config.resp_server_url
        assert d["disklabel_server_url"] == config.disklabel_server_url
        assert d["install_server_dir"] == config.install_server_dir
        assert d["tmp_dir"] == config.tmp_dir



import os
from os.path import isdir, isfile, join, exists


class TestCfg():
    def test_output_path_in_temp_directory(self, cfg, tmpdir): # noqa
        paths = (cfg.output_dir,
                 cfg.sets_output_dir,
                 cfg.resp_output_dir,
                 cfg.img_output_dir,
                 cfg.tmp_dir)

        for p in paths:
            common = os.path.commonpath(paths=(p, str(tmpdir)))
            assert common == str(tmpdir), f"{p} not in {tmpdir}"

    def test_templates_from_fixtures_dir(self, cfg):
        assert "fixtures/config/templates" in cfg.templates_dir
        assert exists(join(cfg.templates_dir, "install.yaml"))


def test_hosts_dir(hosts_dir):
    bravo_host_dir = join(hosts_dir, "bravo")
    assert isdir(bravo_host_dir)


def test_templates_dir(templates_dir):
    template_install_conf = join(templates_dir, "install.conf.j2")
    assert isfile(template_install_conf)


def test_tmpdir(work_dir, tmpdir):
    assert str(tmpdir).startswith(work_dir)


def test_mirror(cfg, mirror_dir2):
    stamp = join(mirror_dir2, ".stamp")
    assert isfile(stamp)
    assert mirror_dir2.startswith(cfg.mirror)



"""Patch installer image with auto_install and site sets."""
import logging
import os
import shutil

from os import makedirs
from os.path import abspath, basename, dirname, exists, join, isfile
from puffmatic import utils

logger = logging.getLogger().getChild(__name__)


def _run(*args, **kwargs):
    kwargs["logger"] = logger
    kwargs["check"] = True
    utils.run(*args, **kwargs)


def mount_img(vnd, img_file, mnt_dir):
    """
    Mount image file using vnd.

    This routine requires priviledged access to vnconfig and mount.
    """
    img_dir = dirname(img_file)
    vnd = f"vnd{vnd}"
    _run(["vnconfig", "-v", vnd, img_file],
         cwd=img_dir,
         priv=True)
    _run(["mount", "-o", "noperm", f"/dev/{vnd}a", mnt_dir],
         cwd=img_dir,
         priv=True)


def umount_img(vnd, mnt_dir):
    """Unmount image attached to vnd."""
    _run(["umount", mnt_dir], priv=True)
    _run(["vnconfig", "-v", "-u", f"vnd{vnd}"], priv=True)


def install_file(src, dst, mode):
    """Install file using install utility."""
    _run(["install", "-m", f"0{mode:o}", src, dst], priv=True)


def decompress_file(input_file, output_file, mode=0o644):
    """Decompress file using gzip."""
    with open(output_file, "wb") as ofile:
        _run(["zcat", input_file], stdout=ofile)
    if mode:
        os.chmod(output_file, mode)


def compress_file(input_file, output_file):
    """Decompress file using gzip."""
    with open(output_file, "wb") as ofile:
        _run(["compress", "-9", "-c", input_file], stdout=ofile)


def extract_root_fs(rd_file, root_fs):
    """Extract root fs from bsd ramdisk image."""
    _run(["rdsetroot", "-x", rd_file, root_fs])


def set_root_fs(rd_file, root_fs):
    """Set root fs in ramdisk kernel."""
    _run(["rdsetroot", rd_file, root_fs])


def patch_root_fs(cfg: utils.Config, disklabel_file, response_file, root_fs_file):
    """Mount root fs image and copy installer config files."""
    root_fs_dir = join(cfg.img_output_dir, "mnt", "rd")
    os.makedirs(root_fs_dir, exist_ok=True)
    mount_img(vnd=2, img_file=root_fs_file, mnt_dir=root_fs_dir)
    target_response_file = join(root_fs_dir, "auto_install.conf")
    target_disklabel_file = join(root_fs_dir, "disklabel")
    install_file(src=response_file, dst=target_response_file, mode=0o644)
    if exists(disklabel_file):
        install_file(src=disklabel_file, dst=target_disklabel_file, mode=0o644)
    umount_img(vnd=2, mnt_dir=root_fs_dir)


def patch_bsd_rd(cfg: utils.Config, host_dir, bsd_rd_file):
    """Patch ramdisk file with autoinstall script and sets."""
    bsd_rd_expanded_file = join(cfg.img_output_dir, "bsd.rd.expanded")
    decompress_file(bsd_rd_file, bsd_rd_expanded_file)
    root_fs_file = join(cfg.img_output_dir, "rd.fs")
    extract_root_fs(bsd_rd_expanded_file, root_fs_file)
    disklabel_file = join(host_dir, "disklabel")
    response_file = join(host_dir, "install.conf")
    patch_root_fs(cfg=cfg,
                  disklabel_file=disklabel_file,
                  response_file=response_file,
                  root_fs_file=root_fs_file)
    set_root_fs(bsd_rd_expanded_file, root_fs_file)
    compress_file(bsd_rd_expanded_file, bsd_rd_file)


def patch_boot_conf(host_dir, boot_conf_file):
    """
    Patch /etc/boot.conf to boot from patched bsd.rd.

    It loads host boot.conf if exists and append bsd.rd
    load stanza.
    """
    host_boot_conf = join(host_dir, "boot.conf")
    boot_conf_lines = []
    if isfile(host_boot_conf):
        with open(host_boot_conf, "r") as f:
            boot_conf_lines = [line.strip() for line in f.readlines()]
    boot_conf_lines.append("set image /bsd.rd")
    with open(boot_conf_file, "w") as of:
        content = "\n".join(boot_conf_lines)
        of.write(content)


def copy_site_sets(cfg: utils.Config, hostname, src_sets_dir, dst_sets_dir):
    """
    Copy site set archives from source to dst directory.

    Files are copied only if site sets exist.
    """
    assert hostname
    assert cfg.version_short != (0, 0)
    host_site_tgz = join(src_sets_dir,
                         f"site{cfg.version_short}-{hostname}.tgz")
    site_tgz = join(src_sets_dir,
                    f"site{cfg.version_short}.tgz")
    if isfile(host_site_tgz):
        logger.info(f"copying {host_site_tgz} to {dst_sets_dir}/")
        makedirs(dst_sets_dir, exist_ok=True)
        shutil.copy(host_site_tgz, dst_sets_dir)
    if isfile(site_tgz):
        logger.info(f"copying {site_tgz} to {dst_sets_dir}/")
        makedirs(dst_sets_dir, exist_ok=True)
        shutil.copy(site_tgz, dst_sets_dir)


def patch_img(cfg: utils.Config, host_dir, install_img_file):
    """
    Patch installXY.img installer image.

    1. mount installer img filesystem
    2. add autoinstall script to bsd.rd
    3. copy site sets to tarballs
    """
    hostname = basename(host_dir)
    img_mnt_dir = join(cfg.img_output_dir, "mnt", "img")
    os.makedirs(img_mnt_dir, exist_ok=True)
    mount_img(vnd=1, img_file=install_img_file, mnt_dir=img_mnt_dir)
    bsd_rd = join(img_mnt_dir, "bsd.rd")
    boot_conf = join(img_mnt_dir, "etc", "boot.conf")
    img_sets_dir = join(img_mnt_dir, cfg.version_full, cfg.arch)
    patch_bsd_rd(cfg=cfg,
                 host_dir=host_dir,
                 bsd_rd_file=bsd_rd)
    patch_boot_conf(host_dir=host_dir,
                    boot_conf_file=boot_conf)
    copy_site_sets(cfg=cfg,
                   hostname=hostname,
                   src_sets_dir=cfg.sets_output_dir,
                   dst_sets_dir=img_sets_dir)
    umount_img(vnd=1,
               mnt_dir=img_mnt_dir)


def patchimg_main(cfg, host_dir, **kwargs):
    """
    Patch installer img file with autoinstall script and site sets.

    It requires upstream installXY.img.
    It copies the installer image adding a hostname suffix.
    Then, host-specific installer is patched.
    """
    host_dir = abspath(host_dir)
    hostname = basename(host_dir)
    orig_img_file = join(cfg.img_output_dir,
                         f"install{cfg.version_short}.img")
    host_img_file = join(cfg.img_output_dir,
                         f"install{cfg.version_short}-{hostname}.img")
    if not exists(host_img_file):
        logger.debug(f"copying {orig_img_file} to {host_img_file}")
        shutil.copy(orig_img_file, host_img_file)
    patch_img(cfg=cfg, host_dir=host_dir, install_img_file=host_img_file)


def patchimg_umount_main(cfg, **kwargs):
    """Unmounts image and ramdisk filesystem images."""
    img_mnt_dir = join(cfg.img_output_dir, "mnt", "img")
    rd_mnt_dir = join(cfg.img_output_dir, "mnt", "rd")
    umount_img(vnd=2, mnt_dir=rd_mnt_dir)
    umount_img(vnd=1, mnt_dir=img_mnt_dir)



from os import makedirs
from os.path import isdir, isfile, join
from puffmatic import getimg
from unittest import mock

import pytest
import shutil
import subprocess


@mock.patch("puffmatic.utils.run")
def test_run(run):
    getimg._run(["alfa", "1", "2"])
    assert run.call_count == 1
    assert run.call_args.args == (["alfa", "1", "2"], )
    assert run.call_args.kwargs.get("logger") is not None
    assert run.call_args.kwargs.get("check") is True


def test_download_file(cfg):
    # GIVEN
    #     configured rsync mirror
    #     configured system version and arch
    # WHEN
    #     downloading single file from mirror
    getimg.download_file(mirror=cfg.mirror,
                         version=cfg.version_full,
                         arch=cfg.arch,
                         file="SHA256",
                         output_dir=cfg.img_output_dir)
    # THEN
    #     output directory created
    #     file stored in output directory
    assert isdir(cfg.img_output_dir), "dir not created"
    assert isfile(join(cfg.img_output_dir, "SHA256"))


class TestCheckSha256():
    def test_passed(self, cfg, mirror_dir2):
        # GIVEN
        #     files from OpenBSD distribution
        #     files are not corrupted
        hash_file = "SHA256"
        img_file = "BUILDINFO"
        assert isfile(join(mirror_dir2, hash_file))
        assert isfile(join(mirror_dir2, img_file))

        # WHEN
        #     integrity check of one file
        getimg.check_sha256(working_dir=mirror_dir2,
                            hash_file="SHA256",
                            img_file="BUILDINFO")
        # THEN
        #     passed

    def test_failed(self, cfg, mirror_dir2, tmpdir):
        # GIVEN
        #     SHA256 containing hashes
        #     a file from OpenBSD distribution
        #     the file is damanged
        shutil.copy(src=join(mirror_dir2, "SHA256"),
                    dst=join(tmpdir, "SHA256"))
        with open(join(tmpdir, "BUILDINFO"), "w") as f:
            f.write("forcing sha256 check failure")

        # WHEN
        #     integrity check
        with pytest.raises(subprocess.CalledProcessError) as e:
            getimg.check_sha256(working_dir=tmpdir,
                                hash_file="SHA256",
                                img_file="BUILDINFO")

        # THEN
        #     corruption detected
        assert e.value.returncode == 1


class TestGetimgMain():

    def test_getimg_main(self, cfg):
        # GIVEN
        #      configured mirror
        #      openbsd version
        #      architecture
        # WHEN
        #      image is retrieved
        getimg.getimg_main(cfg=cfg, once=False)

        # THEN
        #      image and sha files in output dir
        #      files match configured openbsd release version
        assert isfile(join(cfg.img_output_dir, "SHA256"))
        assert isfile(join(cfg.img_output_dir,
                           f"install{cfg.version_short}.img"))

    @mock.patch("puffmatic.getimg.download_file")
    def test_skip_download_if_stamp_exists(self, download_file, cfg):
        # GIVEN
        #     mirror stamp exists
        img_stamp_file = join(cfg.img_output_dir,
                              f"install{cfg.version_short}.img.stamp")
        makedirs(cfg.img_output_dir, exist_ok=True)
        with open(img_stamp_file, "w") as f:
            f.write("test stamp")

        # WHEN
        #     download image with "once" option
        rc = getimg.getimg_main(cfg=cfg, once=True)

        # THEN
        #     download skipped
        assert rc == 0
        assert download_file.call_count == 0



import logging
import os
import shutil
import sys
import pytest

from puffmatic.__main__ import load_config, main
from contextlib import contextmanager
from unittest import mock


@contextmanager
def _chdir(path):
    original_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(original_cwd)


class TestLoadConfig():

    def test_load_config(self, config_file):
        c = load_config(config_file)
        assert c.install_server == "http://install.example.com"

    def test_load_config_from_cwd(self, config_file, tmp_path):
        shutil.copy(config_file, tmp_path)
        with _chdir(tmp_path):
            c = load_config(None)
        assert c.install_server == "http://install.example.com"

    def test_cannot_load_file(self, tmp_path):
        with _chdir(tmp_path):
            with pytest.raises(RuntimeError) as e:
                _ = load_config(None)
        assert str(e.value) == "no config file provided"


class TestMain():

    def setup_method(self):
        self.args = sys.argv

    def teardown_method(self):
        sys.argv = self.args

    def test_help(self, capsys):
        main(["help"])
        captured = capsys.readouterr()
        assert "usage:" in captured.out

    def test_config_load_failed(self):
        # GIVEN
        #     config file cannot be loaded
        args = ["--config", "does-not-exist", "env"]

        # WHEN
        #     run command
        rc = main(args)

        # THEN
        #     failure
        #     root cause logged
        assert rc == 1

    def test_load_config(self, capsys, config_file):
        # GIVEN
        #     valid config
        args = ["--config", config_file, "env"]

        # WHEN
        #     env called
        rc = main(args)

        # THEN
        #     success
        #     config parsed
        #     config printed to stdout
        assert rc == 0
        captured = capsys.readouterr()
        assert "install.example.com" in captured.out

    def test_verbose_mode(self, caplog, config_file):
        # GIVEN
        #     verbose mode enabled
        args = ["--config", config_file, "--verbose", "env"]

        # WHEN
        #     help is printed
        main(args)

        # THEN
        #     logger prints debug message
        assert caplog.records[0].levelno == logging.DEBUG
        assert caplog.records[0].message == "debug is enabled"

    @mock.patch("puffmatic.__main__.mirror_main")
    def test_mirror(self, mirror_main, cfg, config_file):
        args = ["--config", config_file, "mirror"]
        main(args)
        c = mirror_main.call_args.args[0]
        assert c.install_server == cfg.install_server

    @mock.patch("puffmatic.__main__.host_site_set_main")
    def test_host_site_set_main(self, host_site_set_main,
                                cfg, hosts_dir, config_file):
        host_dir = os.path.join(hosts_dir, "bravo")
        main(["--config", config_file, "host-site-set", "-H", host_dir])
        args = host_site_set_main.call_args.args
        kwargs = host_site_set_main.call_args.kwargs
        assert args[0].version_short == "77"
        assert kwargs["host_dir"] == host_dir

    @mock.patch("puffmatic.__main__.site_set_main")
    def test_site_set_main(self, site_set_main,
                           cfg, hosts_dir, config_file):
        site_dir = os.path.join(hosts_dir, "bravo")
        main(["--config", config_file, "site-set", "--site-dir", site_dir])
        args = site_set_main.call_args.args
        kwargs = site_set_main.call_args.kwargs
        assert args[0].version_short == "77"
        assert kwargs["site_dir"] == site_dir

    @mock.patch("puffmatic.__main__.disklabel_main")
    def test_disklabel(self, disklabel_main, hosts_dir, config_file):
        host_dir = os.path.join(hosts_dir, "bravo")
        args = ["--config", config_file, "disklabel", "--host-dir", host_dir]
        main(args)
        kwargs = disklabel_main.call_args.kwargs
        assert host_dir == kwargs["host_dir"]

    @mock.patch("puffmatic.__main__.response_main")
    def test_response(self, response_main, hosts_dir, config_file):
        host_dir = os.path.join(hosts_dir, "bravo")
        args = ["--config", config_file, "response", "--host-dir", host_dir]
        main(args)
        kwargs = response_main.call_args.kwargs
        assert host_dir == kwargs["host_dir"]

    @mock.patch("puffmatic.__main__.getimg_main")
    def test_getimg(self, getimg_main, config_file):
        args = ["--config", config_file, "getimg"]
        main(args)
        args = getimg_main.call_args.args
        assert args[0].version_short == "77"

    @mock.patch("puffmatic.__main__.patchimg_main")
    def test_patchimg(self, patchimg_main, hosts_dir, config_file):
        host_dir = os.path.join(hosts_dir, "bravo")
        args = ["--config", config_file, "patchimg", "--host-dir", host_dir]
        main(args)
        args = patchimg_main.call_args.args
        kwargs = patchimg_main.call_args.kwargs
        assert args[0].version_short == "77"
        assert kwargs["host_dir"] == host_dir

    @mock.patch("puffmatic.__main__.patchimg_umount_main")
    def test_patchimg_umount(self, getimg_umount_main, config_file):
        args = ["--config", config_file, "patchimg-umount"]
        main(args)
        assert getimg_umount_main.call_count == 1



import gzip
import pytest
import shutil

from tempfile import TemporaryDirectory
from os.path import dirname, exists, isfile, join
from os import makedirs
from unittest import mock

from . import fixture_file
from puffmatic.patchimg import compress_file, decompress_file, patch_boot_conf
from puffmatic import patchimg


@mock.patch("puffmatic.utils.run")
def test_mount_img(run, tmpdir):
    mnt_dir = join(tmpdir, "mnt")
    img_file = join(tmpdir, "install.img")
    patchimg.mount_img(vnd=1,
                       img_file=img_file,
                       mnt_dir=mnt_dir)
    assert run.call_count == 2
    vnconfig = run.call_args_list[0]
    assert vnconfig.kwargs["priv"] is True
    assert vnconfig.kwargs["cwd"] == dirname(img_file)
    assert vnconfig.kwargs["check"] is True
    assert vnconfig.kwargs["logger"]

    mount = run.call_args_list[1]
    assert mount.kwargs["priv"] is True
    assert mount.kwargs["cwd"] == dirname(img_file)
    assert mount.kwargs["check"] is True
    assert mount.kwargs["logger"]


@mock.patch("puffmatic.utils.run")
def test_umount_img(run, tmpdir):
    mnt_dir = join(tmpdir, "mnt")
    patchimg.umount_img(vnd=1,
                        mnt_dir=mnt_dir)
    assert run.call_count == 2
    vnconfig = run.call_args_list[0]
    assert vnconfig.kwargs["priv"] is True
    assert vnconfig.kwargs["check"] is True
    assert vnconfig.kwargs["logger"]

    umount = run.call_args_list[1]
    assert umount.kwargs["priv"] is True
    assert umount.kwargs["check"] is True
    assert umount.kwargs["logger"]


@mock.patch("puffmatic.utils.run")
def test_install_file(run):
    src = "/tmp/source.txt"
    dst = "/mnt/target.txt"
    mode = 0o644
    patchimg.install_file(src=src, dst=dst, mode=mode)
    install = run.call_args_list[0]
    assert install.kwargs["priv"] is True
    assert "0644" in install.args[0]


def test_decompress_file():
    input_file = fixture_file("patchimg/sample.txt.gz")
    ref_file = fixture_file("patchimg/sample.txt")

    with TemporaryDirectory() as tmp:
        output_file = join(tmp, "output.txt")
        decompress_file(input_file, output_file)
        with open(output_file, "rb") as f, open(ref_file, "rb") as rf:
            assert f.read() == rf.read()


def test_compress_file():
    input_file = fixture_file("patchimg/sample.txt")

    with TemporaryDirectory() as tmp:
        output_file_gz = join(tmp, "output.txt.gz")
        compress_file(input_file, output_file_gz)
        with (open(input_file, "r") as ref_file,
              gzip.open(output_file_gz, "rb") as compressed_file):
            assert ref_file.read() == compressed_file.read().decode("utf-8")


@mock.patch("puffmatic.utils.run")
def test_extract_root_fs(run):
    rd_file = "/tmp/bsd.rd"
    root_fs = "/tmp/bsd.fs"
    patchimg.extract_root_fs(rd_file=rd_file, root_fs=root_fs)
    install = run.call_args_list[0]
    assert "-x" in install.args[0]
    assert install.kwargs["logger"] is not None
    assert install.kwargs["check"] is True


@mock.patch("puffmatic.utils.run")
def test_set_root_fs(run):
    rd_file = "/tmp/bsd.rd"
    root_fs = "/tmp/bsd.fs"
    patchimg.set_root_fs(rd_file=rd_file, root_fs=root_fs)
    install = run.call_args_list[0]
    assert "-x" not in install.args[0]
    assert install.kwargs["logger"] is not None
    assert install.kwargs["check"] is True


class TestPatchBootConf():

    @pytest.fixture
    def boot_conf_file(self, tmpdir):
        return tmpdir / "boot.conf"

    @pytest.fixture
    def host_dir(self):
        boot_conf = fixture_file("patchimg/boot.conf")
        return dirname(boot_conf)

    def test_image_appended_to_host_boot_conf(self, host_dir, boot_conf_file):
        patch_boot_conf(host_dir, boot_conf_file=boot_conf_file)
        with open(boot_conf_file, "r") as f:
            lines = [line.strip() for line in f.readlines()]
        assert "stty com0 115200" in lines
        assert "set tty com0" in lines
        assert "set image /bsd.rd" == lines[-1]

    def test_no_host_boot_conf(self, tmpdir, boot_conf_file):
        empty = tmpdir / "empty"
        makedirs(empty)
        patch_boot_conf(host_dir=empty, boot_conf_file=boot_conf_file)
        with open(boot_conf_file, "r") as f:
            lines = [line.strip() for line in f.readlines()]
        assert ["set image /bsd.rd"] == lines


def test_copy_site_sets(cfg, tmp_path):
    # GIVEN
    #     general site set
    #     host specific site set
    src_dir = join(tmp_path, "src_sets")
    makedirs(src_dir)
    sets = [f"site{cfg.version_short}.tgz",
            f"site{cfg.version_short}-{cfg.hostname}.tgz"]
    for s in sets:
        spath = join(src_dir, s)
        with open(spath, "w") as f:
            f.write("dummy site set archive")

    # WHEN
    #     sets copied to destination directory
    patchimg.copy_site_sets(cfg=cfg,
                            hostname=cfg.hostname,
                            src_sets_dir=src_dir,
                            dst_sets_dir=cfg.sets_output_dir)
    # THEN
    #     both sets copied
    for s in sets:
        spath = join(cfg.sets_output_dir, s)
        assert isfile(spath)


class TestPatchImgMain():

    def teardown_method(self, *args):
        """
        Unmount images and deactive vnd devices.

        Under normal operation vnd should be cleaned by the code under
        test. However, if the test fails, cleanup is prevented by
        exception. Attempt cleanup first (skipping errors), then
        verify if vnd devices are deactived.
        """
        from subprocess import run
        for vnd in ["vnd1", "vnd2"]:
            run(["doas", "umount", f"/dev/{vnd}"], check=False)
            run(["doas", "vnconfig", "-u", vnd], check=False)
        for vnd in ("vnd1", "vnd2"):
            vnd_chk = run(["doas", "vnconfig", "-l", vnd],
                          capture_output=True).stdout
            assert f"{vnd}: not in use" in str(vnd_chk), f"/dev/{vnd} active"

    def test_patchimg_main(self, cfg, hosts_dir, mirror_dir2):
        # GIVEN
        #     installXY.img downloaded into img directory
        #     host configuration present
        mirror_install_img = join(mirror_dir2,
                                  f"install{cfg.version_short}.img")
        assert isfile(mirror_install_img)
        host_dir = join(hosts_dir, "example")
        orig_install_img = join(cfg.img_output_dir,
                                f"install{cfg.version_short}.img")
        makedirs(cfg.img_output_dir)
        shutil.copy(mirror_install_img, orig_install_img)

        # WHEN
        #     patching installXY.img
        patchimg.patchimg_main(cfg=cfg,
                               host_dir=host_dir)

        # THEN
        #     original image not changed
        #     patched host specific image created
        patched_install_img = join(cfg.img_output_dir,
                                   f"install{cfg.version_short}-example.img")
        assert exists(patched_install_img)


@mock.patch("puffmatic.patchimg.umount_img")
def test_patchimg_umount_main(umount_img, cfg):
    # WHEN
    #     umount cleanup called
    patchimg.patchimg_umount_main(cfg)
    # THEN
    #     vnds are unmounted in reverse order
    first = umount_img.call_args_list[0].kwargs["vnd"]
    second = umount_img.call_args_list[1].kwargs["vnd"]
    assert first > second



import os

__dir = os.path.dirname(__file__)
fixtures_dir = os.path.join(__dir, "fixtures")


def fixture_file(p):
    """Return fixture file, if exists."""
    file_path = os.path.join(fixtures_dir, p)
    if not os.path.exists(file_path):
        raise AssertionError(f"no {p} in fixture")
    return file_path



"""Various utility helpers."""
import dataclasses
import inspect
import logging
import select
import subprocess
import sys
import threading
import urllib
import yaml
from dataclasses import (dataclass, field)
from os.path import join, abspath, basename, exists


def create_logger(name=None, level=logging.INFO):
    """Create logger instance."""
    logger = logging.getLogger(name)
    logger.setLevel(level)
    console_handler = logging.StreamHandler(stream=sys.stdout)
    formatter = logging.Formatter(
        fmt="{asctime} {levelname} {name}: {message}",
        style="{",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    console_handler.setFormatter(formatter)
    console_handler.setLevel(logging.DEBUG)
    logger.addHandler(console_handler)
    return logger


def _cmd_output_logger(logger: logging.Logger = None, stdout=None, stderr=None):
    """Redirects stdout and stderr streams to logger."""
    if logger is None:
        return
    streams = [s for s in (stdout, stderr) if s is not None]
    while streams:
        ready_to_read, _, _ = select.select(streams, [], [])
        for stream in ready_to_read:
            line = stream.readline()
            if line:
                logger.info(line.rstrip())
            else:
                streams.remove(stream)


def _format_cmd(cmd, stdout=None, stderr=None, stdin=None):
    """Format shell command, including stream redirection operators."""
    c = cmd.copy()
    if hasattr(stdin, "name") and exists(stdin.name):
        c.extend(("<", stdin.name))
    if hasattr(stdout, "name") and exists(stdout.name):
        c.extend((">", stdout.name))
    if hasattr(stderr, "name") and exists(stderr.name):
        c.extend(("2>", stderr.name))
    return " ".join(c)


def run(cmd, **kwargs):
    """Run shell command redirecting output to logger."""
    stdout = kwargs.pop("stdout", subprocess.PIPE)
    stderr = kwargs.pop("stderr", subprocess.PIPE)
    parent_logger = kwargs.pop("logger", None)
    text = kwargs.pop("text", True)
    priv = kwargs.pop("priv", False)
    check = kwargs.pop("check", False)

    full_cmd = cmd if not priv else ["doas"] + cmd
    process = subprocess.Popen(
        full_cmd,
        stdout=stdout,
        stderr=stderr,
        text=text,
        **kwargs
    )
    reader = None
    if parent_logger:
        suffix = basename(cmd[0]).replace(".", "_")
        logger = parent_logger.getChild(suffix)
        output_logger_args = {
            "logger": logger,
            "stdout": process.stdout if stdout == subprocess.PIPE else None,
            "stderr": process.stderr if stderr == subprocess.PIPE else None,
        }
        reader = threading.Thread(target=_cmd_output_logger,
                                  kwargs=output_logger_args,
                                  daemon=True)
        reader.start()
        formatted_cmd = _format_cmd(full_cmd,
                                    stdout=stdout,
                                    stderr=stderr,
                                    stdin=kwargs.get("stdin", None))
        parent_logger.debug(f"running {formatted_cmd}")
    rc = process.wait()
    if reader:
        reader.join()
    if check and rc:
        raise subprocess.CalledProcessError(rc, full_cmd)
    return rc


def parse_version(version: str | tuple) -> (int, int):
    """Parse version and return tuple of (major, minor)."""
    match version:
        case (int() as major, int() as minor):
            return (major, minor)
    v = version.split(".")
    if len(v) != 2:
        return None
    try:
        major = int(v[0])
        minor = int(v[1])
        return major, minor
    except ValueError:
        return None


def load_yaml(*files):
    """Load all yaml files and merge them together."""
    merged_data = {}
    for file in files:
        with open(file) as f:
            data = yaml.safe_load(f)
            merged_data.update(data)
    return merged_data


@dataclass
class Config():
    """
    Config provides global settings for all subcommands.

    It provides information about all locations as well
    as values for config file template rendering.
    """

    version: (int, int) = field(default=(0, 0))
    arch: str = field(default=None)
    mirror: str = field(default=None)
    install_server: str = field(default=None)
    install_server_dir: str = field(default=None)
    install_basic_auth: str = field(default=None)
    templates_dir: str = field(default=None)
    output_dir: str = field(default=None)

    hostname: str = field(default=None)
    ssh_public_key: str = field(default=None)
    root_password: str = field(default=None)
    disk_encryption_passphrase: str = field(default=None)
    disklabel: str = field(default=None)
    ipv6_address: str = field(default=None)
    ipv6_gateway: str = field(default=None)
    use_autopartitioning_template: bool = field(default=False)

    def __post_init__(self):
        """Convert values to native types."""
        self.version = parse_version(self.version)
        if self.templates_dir:
            self.templates_dir = abspath(self.templates_dir)
        if self.output_dir:
            self.output_dir = abspath(self.output_dir)

    @property
    def version_full(self):
        """Format version."""
        return f"{self.version[0]}.{self.version[1]}"

    @property
    def version_short(self):
        """Format version in short format, without dot."""
        return f"{self.version[0]}{self.version[1]}"

    @property
    def sets_output_dir(self):
        """Directory where output site tests are written."""
        return join(self.output_dir, "sets")

    @property
    def resp_output_dir(self):
        """Directory where output response files are written."""
        return join(self.output_dir, "resp")

    @property
    def img_output_dir(self):
        """Directory where output installer image file is written."""
        return join(self.output_dir, "img")

    @property
    def tmp_dir(self):
        """Temporary directory."""
        return join(self.output_dir, "tmp")

    @property
    def resp_server_url(self):
        """Formatted server url where response file can be downloaded."""
        u = urllib.parse.urlparse(self.install_server)
        if self.install_basic_auth:
            return f"{u.scheme}://{self.install_basic_auth}@{u.netloc}"
        else:
            return f"{u.scheme}://{u.netloc}"

    @property
    def disklabel_server_url(self):
        """Formatted server url where disklabel file can be downloaded."""
        return self.resp_server_url

    @property
    def install_server_url(self):
        """Formatted server url where installation sets are hosted."""
        return self.install_server

    @classmethod
    def load(cls, *files: str):
        """Load environment configuration."""
        return Config().update(*files)

    def update(self, *files, **kwargs):
        """Upload environment from set of files."""
        d = load_yaml(*files)
        d.update(kwargs)
        return dataclasses.replace(self, **d)

    def to_dict(self):
        """Convert config objects to dictionary."""
        d = dataclasses.asdict(self)
        props = inspect.getmembers(self.__class__,
                                   lambda p: isinstance(p, property))
        for name, attr in props:
            d[name] = getattr(self, name)
        return d



"""Download installer image files."""
import logging
import os

from puffmatic import utils
from datetime import datetime, timezone


logger = logging.getLogger().getChild(__name__)


def _run(*args, **kwargs):
    kwargs["logger"] = logger
    kwargs["check"] = True
    utils.run(*args, **kwargs)


def download_file(mirror, version, arch, file, output_dir):
    """Download a file from OpenBSD rsync mirror."""
    os.makedirs(output_dir, exist_ok=True)
    _run(["openrsync", "-v", f"{mirror}/{version}/{arch}/{file}", "./"],
         cwd=output_dir)


def check_sha256(working_dir, hash_file, img_file):
    """Check installer image by running hash check."""
    _run(["sha256", "-C", hash_file, img_file], cwd=working_dir)


def getimg_main(cfg, once, **kwargs):
    """
    Get installer image subcommand.

    It downloads OpenBSD USB installer image and places it in output
    directory. SHA256 is verified as well.
    """
    img_file = f"install{cfg.version_short}.img"
    sha_file = "SHA256"
    img_stamp_file = os.path.join(cfg.img_output_dir,
                                  f"{img_file}.stamp")
    if once and os.path.isfile(img_stamp_file):
        logger.info(f"found {img_stamp_file} - skipping")
        return 0
    for file in [img_file, "SHA256"]:
        download_file(mirror=cfg.mirror,
                      version=cfg.version_full,
                      arch=cfg.arch,
                      file=file,
                      output_dir=cfg.img_output_dir)
    check_sha256(working_dir=cfg.img_output_dir,
                 hash_file=sha_file,
                 img_file=img_file)
    with open(img_stamp_file, "w") as f:
        ts = datetime.now(tz=timezone.utc)
        f.write(f"downloaded from {cfg.mirror} at {ts}")



import pytest
import filecmp

from os.path import join, isfile
from puffmatic.disklabel import disklabel_main


class TestDisklabelMain():

    def test_copy_host_disklabel_file(self, cfg, hosts_dir):
        # GIVEN
        #     host config has custom disklabel
        host = join(hosts_dir, "example")
        src_disklabel = join(host, "disklabel")
        dst_disklabel = join(cfg.resp_output_dir, "disklabel-example")

        # WHEN
        #     disklabel is copied
        disklabel_main(cfg, host)

        # THEN
        #     disklabel is copied from host dir to output dir
        #     disklabel has host name suffix
        assert isfile(dst_disklabel)
        assert "-example" in dst_disklabel
        assert filecmp.cmp(src_disklabel, dst_disklabel)

    def test_copy_generic_disklabel_file(self, cfg, tmpdir):
        # GIVEN
        #     host has no disklabel
        #     generic disklabel exists
        host = join(tmpdir, "labelless")
        src_disklabel = join(cfg.templates_dir, "disklabel")
        dst_disklabel = join(cfg.resp_output_dir, "disklabel-labelless")

        # WHEN
        #     disklabel is copied
        disklabel_main(cfg, host)

        # THEN
        #     generic disklabel is copied
        #     file has hostname suffix
        assert isfile(dst_disklabel)
        assert "-labelless" in dst_disklabel
        assert filecmp.cmp(src_disklabel, dst_disklabel)

    def test_no_disklabel_found(self, cfg, tmpdir):
        # WHEN
        #     host has no disklabel
        #     no generic disklabel
        host = join(tmpdir, "non-existing")
        cfg.templates_dir = join(tmpdir, "non-existing")

        # WHEN
        #     disklabel copied
        with pytest.raises(SystemExit) as se:
            disklabel_main(cfg, host)

        # THEN
        #     exit with error
        assert se.value.code == 1



"""Utilities to generate response file."""
import jinja2
import logging
import os
import shutil
import sys

from puffmatic import utils
from os import makedirs
from os.path import (abspath,
                     basename,
                     exists,
                     dirname,
                     join)


logger = logging.getLogger().getChild(__name__)


def _run(*args, **kwargs):
    kwargs["logger"] = logger
    kwargs["check"] = True
    utils.run(*args, **kwargs)


def render_install_resp(install_resp_template: str,
                        output_file: str,
                        context):
    """
    Render host response file from template and context.

    File is passed through Jinja only if it ends with .j2 extension.
    Otherwise file is copied verbatim.
    """
    if not install_resp_template.endswith(".j2"):
        shutil.copy(install_resp_template, output_file)
    else:
        template_dir = dirname(install_resp_template)
        template_name = basename(install_resp_template)
        loader = jinja2.FileSystemLoader(searchpath=template_dir)
        env = jinja2.Environment(loader=loader, keep_trailing_newline=True)
        template = env.get_template(template_name)
        os.makedirs(dirname(output_file), exist_ok=True)
        with open(output_file, "w") as f:
            txt = template.render(**context)
            f.write(txt)


def response_main(cfg, host_dir, **kwargs):
    """Render autoinstall response file."""
    host_dir = abspath(host_dir)
    templates_dir = abspath(cfg.templates_dir)
    output_dir = abspath(cfg.resp_output_dir)

    makedirs(output_dir, exist_ok=True)
    hostname = basename(host_dir)
    output_file = join(output_dir, f"install-{hostname}.conf")
    host_install_resp = join(host_dir, "install.conf")
    host_install_resp_template = join(host_dir, "install.conf.j2")
    host_install_resp_config = join(host_dir, "install.yaml")
    generic_install_resp_template = join(templates_dir, "install.conf.j2")
    generic_install_resp_config = join(templates_dir, "install.yaml")

    resp_configs = filter(exists, [generic_install_resp_config,
                                   host_install_resp_config])

    if exists(host_install_resp):
        logging.info(f"copying host response file {host_install_resp} to {output_file}")
        shutil.copy(host_install_resp, output_file)
    elif exists(host_install_resp_template):
        logging.info(f"generating host response file {output_file}")
        config = cfg.update(*resp_configs,
                            hostname=hostname)
        render_install_resp(host_install_resp_template,
                            output_file,
                            config.to_dict())
    elif exists(generic_install_resp_template):
        logging.info(f"generating generic response file {output_file}")
        config = cfg.update(*resp_configs,
                            hostname=hostname)
        render_install_resp(generic_install_resp_template,
                            output_file,
                            config.to_dict())
    else:
        logging.error("no install response file or template")
        sys.exit(1)



"""Mirror module provides OpenBSD pub release directory."""

import logging
import os

from datetime import datetime
from datetime import timezone
from puffmatic import utils

logger = logging.getLogger().getChild(__name__)


def _run(*args, **kwargs):
    kwargs["logger"] = logger
    kwargs["check"] = True
    return utils.run(*args, **kwargs)


def mirror_release_dir(version, arch, output_dir, mirror):
    """Mirror OpenBSD release directory to output_dir."""
    os.makedirs(output_dir, exist_ok=True)
    _run(["openrsync", "-v", "-r", f"{mirror}/{version}/{arch}/", "./"],
         cwd=output_dir)


def check_release_dir(release_dir):
    """Check release directory by running hash check."""
    return _run(["sha256",  "-c", "SHA256"], cwd=release_dir)


def mirror_main(cfg, once, **kwargs):
    """
    Mirror subcommand.

    It downloads OpenBSD release files and places them in output
    directory. Mirror SHA256 is verified as well.
    """
    stamp_file = os.path.join(cfg.sets_output_dir, "mirror.stamp")
    if once and os.path.isfile(stamp_file):
        logger.info(f"found {stamp_file} - skipping")
        return 0

    mirror_release_dir(version=cfg.version_full,
                       arch=cfg.arch,
                       output_dir=cfg.sets_output_dir,
                       mirror=cfg.mirror)
    check_release_dir(cfg.sets_output_dir)
    with open(stamp_file, "w") as f:
        ts = datetime.now(tz=timezone.utc)
        f.write(f"downloaded from {cfg.mirror} at {ts}")



"""Site set generator."""

import logging
import os

from puffmatic import utils
from os.path import (abspath,
                     basename,
                     isdir,
                     join)
import shutil
import sys


logger = logging.getLogger().getChild(__name__)


def _run(*args, **kwargs):
    kwargs["logger"] = logger
    kwargs["check"] = True
    utils.run(*args, **kwargs)


def mount_tmp_dir(tmp_dir, size=64):
    """Mount temporary directory where sets can be created."""
    _run(["mount", "-t", "mfs", "-o", f"noperm,-s{size}M", "swap", tmp_dir],
         priv=True)


def umount_tmp_dir(tmp_dir):
    """Umount temporary directory."""
    _run(["umount", "-f", str(tmp_dir)], priv=True)


def run_mtree(site_dir, mtree_file):
    """Run mtree to fix any site set files deviations."""
    _run(["mtree", "-f", mtree_file, "-U"],
         cwd=site_dir)


def create_tarball(site_dir, output_file):
    """Create site set tarball."""
    _run(["tar", "-c", "-z", "-f", output_file, "."],
         cwd=site_dir)


def update_index(dir):
    """Update index.txt using bsd ls command."""
    index = os.path.join(dir, "index.txt")
    with open(index, "wb") as f:
        _run(["ls", "-l"],
             cwd=dir,
             stdout=f)


def create_site_set(site_dir, mtree_file, output_dir, archive, tmp_dir):
    """Create siteXY.tgz or siteXY-$host.tgz file."""
    site_archive = join(output_dir, archive)
    site_dir_copy = join(tmp_dir, basename(site_dir))
    shutil.copytree(src=site_dir,
                    dst=site_dir_copy)
    run_mtree(site_dir_copy, mtree_file)
    logger.info(f"creating {site_archive}")
    os.makedirs(output_dir, exist_ok=True)
    create_tarball(site_dir_copy, site_archive)
    logger.info(f"updating index in {output_dir}")
    update_index(output_dir)


def host_site_set_main(cfg, host_dir, **kwargs):
    """Generate siteXY-{hostname}.tgz tarballs."""
    host_dir = abspath(host_dir)
    hostname = basename(host_dir)
    archive = f"site{cfg.version_short}-{hostname}.tgz"
    host_site_dir = join(host_dir, "site")
    host_mtree_file = join(host_dir, "site.mtree")
    sets_output_dir = abspath(cfg.sets_output_dir)

    if not isdir(host_site_dir):
        logger.info(f"no site for host {hostname} - skipping")
        sys.exit(0)

    os.makedirs(cfg.sets_output_dir, exist_ok=True)
    os.makedirs(cfg.tmp_dir, exist_ok=True)
    mount_tmp_dir(cfg.tmp_dir, size=32)
    create_site_set(site_dir=host_site_dir,
                    mtree_file=host_mtree_file,
                    output_dir=sets_output_dir,
                    archive=archive,
                    tmp_dir=cfg.tmp_dir)
    umount_tmp_dir(cfg.tmp_dir)


def site_set_main(cfg, site_dir, **kwargs):
    """Generate siteXY.tgz tarball."""
    site_dir = abspath(site_dir)
    mtree_file = abspath(join(site_dir, "..", "site.mtree"))
    sets_output_dir = abspath(cfg.sets_output_dir)
    archive = f"site{cfg.version_short}.tgz"
    os.makedirs(cfg.sets_output_dir, exist_ok=True)
    os.makedirs(cfg.tmp_dir, exist_ok=True)
    mount_tmp_dir(cfg.tmp_dir, size=32)
    create_site_set(site_dir=site_dir,
                    mtree_file=mtree_file,
                    output_dir=sets_output_dir,
                    archive=archive,
                    tmp_dir=cfg.tmp_dir)
    umount_tmp_dir(cfg.tmp_dir)
